from __future__ import annotations

from typing import TYPE_CHECKING

from iminuit import Minuit
from numpy import array, ascontiguousarray

from .fit_result import FitResult
from .minimizer_base import MinimizerBase

if TYPE_CHECKING:
    from dag_modelling.core.output import Output
    from dag_modelling.parameters import Parameter
    from numpy.typing import NDArray

# if we cannot import runtime_error from root we use DagModellingError to avoid any exception capture,
# i.e., if CppRuntimeError==DagModellingError, the exception will be not raised
try:
    import ROOT  # fmt: skip
    CppRuntimeError = ROOT.std.runtime_error
except Exception:
    from dag_modelling.core.exception import DagModellingError  # fmt:skip
    CppRuntimeError = DagModellingError


class IMinuitMinimizer(MinimizerBase):
    __slots__ = ("_errordef",)
    _errordef: float

    def __init__(
        self,
        statistic: Output,
        parameters: dict[str, Parameter],
        name: str = "iminuit",
        label: str = "iminuit",
        errordef: float = 1.0,  # 1.0: LEAST_SQUARES, 0.5: LIKELIHOOD
        *,
        limits: dict[str, tuple[float | None, float | None]] = {},
        **kwargs,
    ) -> None:
        super().__init__(statistic, parameters, name, label, limits=limits, **kwargs)
        self._errordef = errordef

    def _child_fit(self, **kwargs) -> dict:
        """Run Migrad minimization.

        Migrad from the Minuit2 library is a robust minimization
        algorithm, which uses first and approximate second derivatives
        to achieve quadratic convergence near the minimum.
        """
        ncall = kwargs.pop("ncall", None)  # maximum number of calls inside migrad
        iterate = kwargs.pop("iterate", 5)  # N calls if convergence was not reached; default: 5

        result = self.init_minimizer()
        fmin = None
        with FitResult() as fr:
            try:
                result = self._minimizer.migrad(ncall=ncall, iterate=iterate)
            except CppRuntimeError as exc:
                message = f"{exc.what()}"  # pyright: ignore
            except RuntimeError as exc:
                message = repr(exc)
            else:
                # the message by default has str, html and pretty representations,
                # but we build a dict using slots with complete information
                fmin = result.fmin
                message = {key[1:]: getattr(fmin, key) for key in fmin.__slots__}
        fr.set(
            x=array(result.values),
            errors=array(result.errors),
            fun=float(result.fval) if result.fval is not None else None,
            success=result.valid,
            summary=message,
            minimizer=self._label,
            nfev=result.nfcn,
            errorsdef=self._errordef,
            covariance=(array(result.covariance) if result.covariance is not None else None),
            nbins=self.nbins,
            npars_free=self.npars_free,
            npars_constrained=self.npars_constrained,
            ndof=self.nbins - self.npars_free,
        )
        self._result = fr.result
        self.patchresult()

        return self.result

    def init_minimizer(self) -> Minuit:
        """Initializes the Minuit minimizer."""
        if not self._parameters:
            raise RuntimeError("Pass parameters to minimize!")
        minimizable = self.init_minimizable()

        def fcn(*params):
            params = ascontiguousarray(params, dtype="d")
            return minimizable(params)

        startvalues = []
        names = self._parameters_names
        for par in self._parameters:
            # TODO: wrap path getter into Parameter method
            startvalues.append(par.value)

        self._minimizer = minimizer = Minuit(fcn, *startvalues, name=names)
        minimizer.throw_nan = True
        minimizer.errordef = self._errordef

        limits = []
        for name in names:
            if name in self._limits.keys():
                limits.append(self._limits[name])
            else:
                limits.append((None, None))
        self._minimizer.limits = limits

        return minimizer

    def profile_errors(
        self,
        names: list[str] | None = None,
        confidence_level: float | None = None,
        ncall: int | None = None,
    ) -> dict:
        """Calculates errors for parameters within the Minos algorithm.

        The Minos algorithm uses the profile likelihood method to
        compute (generally asymmetric) confidence intervals. It scans
        the negative log-likelihood or (equivalently) the least-squares
        cost function around the minimum to construct a confidence
        interval.
        """
        result = {}
        if names:
            self.logger.info(f"Caclulating profile for: {names}")
            _names = result["names"] = names
        else:
            _names = result["names"] = self.result["names"]
        errs = result["errors"] = []
        errsdict = result["errorsdict"] = {}
        statuses = result["errors_profile_status"] = {}

        for name in _names:
            status = statuses[name] = {}
            try:
                res = self._minimizer.minos(name, cl=confidence_level, ncall=ncall)
                stat = res.merrors[name]
                errs.append([stat.lower, stat.upper])
                errsdict[name] = errs[-1]
                for key in stat.__slots__:
                    status[key] = getattr(stat, key)
                status["message"] = ""
            except CppRuntimeError as exc:
                status["is_valid"] = False
                status["message"] = f"{exc.what()}"  # pyright: ignore
            except RuntimeError as exc:
                status["is_valid"] = False
                status["message"] = repr(exc)

        return result

    def calculate_covariance(self, ncall: int | None = None) -> "NDArray":
        """Calculates covariance matrix within the Hesse algorithm.

        The Hesse method estimates the covariance matrix by inverting the matrix of second derivatives
        (Hesse matrix) at the minimum. To get parameters correlations, you need to use this.
        The Minos algorithm is another way to estimate parameter uncertainties, see `profile_errors()`.

        .. note:: by default the covariance matrix is already calculated and stored in `result["covariance"]`
        """
        res = self._minimizer.hesse(ncall=ncall)
        return res.covariance

    def get_scans(self, names: list, fitresult: dict):
        """Get Minos profile over a specified interval.

        Scans over one parameter and minimises the function with respect
        to all other parameters for each scan point.
        """
        scans = fitresult["scan"] = {}
        if names:
            self.logger.info(f"Caclulating mnprofile for: {names}")
        for name in names:
            scan = scans[name] = {}
            try:
                xout, yout, valid = self._minimizer.mnprofile(name)
            except CppRuntimeError as exc:
                self._on_exception_in_get_scans(scan, f"{exc.what()}")  # pyright: ignore
            except RuntimeError as exc:
                self._on_exception_in_get_scans(scan, repr(exc))
            else:
                scan["x"] = xout.tolist()
                scan["y"] = yout.tolist()
                scan["success"] = valid.tolist()
                scan["message"] = ""

    def _on_exception_in_get_scans(self, scan: dict, msg: str):
        scan["x"] = []
        scan["y"] = []
        scan["success"] = False
        scan["message"] = msg



from math import sqrt

from dag_modelling.lib.statistics import Chi2, CNPStat, MonteCarlo
from dag_modelling.core.graph import Graph
from dag_modelling.core.input import Input
from dag_modelling.lib.abstract import OneToOneNode
from dag_modelling.lib.common import Array
from dag_modelling.parameters import Parameters
from dag_modelling.plot.graphviz import savegraph
from dag_modelling.plot.plot import plot_array_1d
from matplotlib import pyplot as plt
from numpy import allclose, array, linspace
from numpy.random import MT19937, Generator, SeedSequence
from pytest import mark
from scipy.stats import norm

from dgm_fit.iminuit_minimizer import IMinuitMinimizer

_NevScale = 10000


class Model(OneToOneNode):
    __slots__ = ("_mu", "_sigma", "_const")
    _mu: Input
    _sigma: Input
    _const: Input

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._mu = self._add_input("mu", positional=False)
        self._sigma = self._add_input("sigma", positional=False)
        self._const = self._add_input("const", positional=False)

    def _function(self):
        mu = self._mu.data[0]
        sigma = self._sigma.data[0]
        const = self._const.data[0]
        for indata, outdata in zip(self.inputs.iter_data(), self.outputs.iter_data_unsafe()):
            outdata[:] = _NevScale * norm.pdf(indata[:], loc=mu, scale=sigma) + const


@mark.parametrize("corr", (False, True))
@mark.parametrize(
    "mu,sigma,const,mu_limits",
    (
        (-1.531654, 0.567543, 10, None),
        (2.097123, 1.503321, 20, (-5, None)),
        (-1.531654, 0.567543, 30, (None, 5)),
        (2.097123, 1.503321, 40, (-5, 5)),
    ),
)
@mark.parametrize("mode", ("asimov", "normal-stats"))
@mark.parametrize("verbose", (False, True))
def test_IMinuitMinimizer(corr, mu, sigma, const, mu_limits, mode, verbose: bool, testname):
    size = 201
    x = linspace(-10, 10, size)

    # start values of the fitting
    mu_fit = mu - 0.3 * sigma
    sigma_fit = sigma * 1.1
    const_fit = const * 1.1
    with Graph(close_on_exit=True) as graph:
        # setting of true parameters
        Mu0 = Array("mu 0", [mu], mode="fill")
        Sigma0 = Array("sigma 0", [sigma], mode="fill")
        Const0 = Array("const 0", [const], mode="fill")
        X = Array("x", x, mode="fill")

        # build input data for the MC simulation
        pdf0 = Model("normal pdf for MC")
        X >> pdf0
        Mu0 >> pdf0("mu")
        Sigma0 >> pdf0("sigma")
        Const0 >> pdf0("const")
        model = pdf0.outputs[0]

        # perform fluctuations of data within MC and shift the result with constant background
        (sequence,) = SeedSequence(0).spawn(1)
        gen = Generator(MT19937(sequence))
        mc = MonteCarlo("MC", mode=mode, generator=gen)
        model >> mc

        # build a model to fit exp data
        pars = Parameters.from_numbers(
            [mu_fit, sigma_fit, const_fit],
            names=["mu_fit", "sigma_fit", "const_fit"],
            sigma=[1, 1, 1],
            correlation=[[1, -0.95, 0], [-0.95, 1, 0], [0, 0, 1]] if corr else None,
        )
        MuFit, SigmaFit, ConstFit = pars.outputs()
        # MuFit = Array("mu fit", [mufit], mode="fill")
        # SigmaFit = Array("sigma fit", [sigmafit], mode="fill")
        pdf_fit = Model("normal pdf for the Model")
        X >> pdf_fit
        MuFit >> pdf_fit("mu")
        SigmaFit >> pdf_fit("sigma")
        ConstFit >> pdf_fit("const")

        # eval errors
        cnp = CNPStat("CNP stat")
        (mc, pdf_fit) >> cnp

        # eval Chi2
        chi = Chi2("Chi2")
        mc >> chi("data")
        pdf_fit >> chi("theory")
        cnp.outputs[0] >> chi("errors")

    # check if the MC data is valid: negative events -> wrong model
    assert min(mc.outputs[0].data) >= 0

    limits = {}
    if mu_limits:
        limits["mu"] = mu_limits

    # perform a minimization
    par_mu, par_sigma, par_const = pars.parameters
    minimizer = IMinuitMinimizer(
        statistic=chi.outputs[0],
        parameters={
            "mu": par_mu,
            "sigma": par_sigma,
            "const": par_const,
        },
        limits=limits,
        verbose=verbose,
    )
    res = minimizer.fit()

    assert res["success"]
    assert res["nfev"] > 1

    names = res["names"]
    assert (
        len(res["x"])
        == len(names)
        == len(res["errorsdict"])
        == len(res["errors"])
        == len(res["xdict"])
        == res["npars"]
        == 3
    )

    atol = 2.0 / sqrt(_NevScale)
    res_fit = [mu, sigma, const]
    assert allclose(
        res["x"][:2],
        res_fit[:2],
        rtol=0,
        atol=atol if mode == "normal-stats" else 2e-5,
    )
    assert allclose(
        res["x"][2:],
        res_fit[2:],
        rtol=0,
        atol=0.55 if mode == "normal-stats" else 4.0e-3,
    )
    rel_dev = (res["x"] - res_fit) / res["errors"]
    assert allclose(
        rel_dev,
        0,
        rtol=0,
        atol=2 if mode == "normal-stats" else 1.0e-2,
    )
    if mode == "asimov":
        assert res["fun"] < 0.1
    assert allclose(
        res["covariance"],
        minimizer.calculate_covariance(),
        rtol=0,
        atol=mode == "asimov" and 1e-6 or 1.0e-5,
    )
    assert all(res["errorsdict"][key] == res["errors"][i] for i, key in enumerate(names))

    # errors checks
    errors = minimizer.profile_errors()
    assert errors["names"] == names
    errs = array(errors["errors"])
    assert allclose(errs[:, 1], res["errors"], atol=4e-3)
    assert allclose(-errs[:, 0], res["errors"], atol=4e-3)
    assert all((errors["errorsdict"][key] == errs[i]).all() for i, key in enumerate(names))
    for name in names:
        for key in ("is_valid", "lower_valid", "upper_valid"):
            assert errors["errors_profile_status"][name][key]
        assert errors["errors_profile_status"][name]["message"] == ""

    # save plot and graph
    draw_params(res["x"], mu, sigma, minimizer, f"output/{testname}-params.png")
    draw_fit(x, mc, model, pdf_fit.outputs[0], mode, f"output/{testname}-plot.png")
    savegraph(graph, f"output/{testname}.png")


def draw_params(res, mu, sigma, minimizer, figname):
    _, ax = plt.subplots()
    for cl in (1, 2, 3):
        contours = minimizer._minimizer.mncontour(0, 1, cl=cl)
        ax.plot(contours[:, 0], contours[:, 1], label=f"{cl}σ")
    ax.scatter(*[mu, sigma], label="true")
    ax.scatter(*res, label="fit")
    plt.xlabel("mu")
    plt.ylabel("sigma")
    plt.legend()
    plt.xlim([mu - 0.01, mu + 0.01])
    plt.ylim([sigma - 0.01, sigma + 0.01])
    plt.savefig(figname)
    print("Write", figname)
    plt.close()


def draw_fit(x, mc, model, modelfit, mode, figname):
    ax = plt.subplot(111)
    ax.minorticks_on()
    ax.grid()
    ax.set_xlabel("x")
    ax.set_ylabel("y")
    plot_array_1d(
        mc.outputs[0].data,
        meshes=x,
        color="black",
        label="data+fluct." if mode != "asimov" else "asimov MC",
    )
    plot_array_1d(model.data, meshes=x, linestyle="--", label="data")
    plot_array_1d(modelfit.data, meshes=x, linestyle="--", label="fit")
    ax.legend()
    plt.savefig(figname)
    print("Write", figname)
    plt.close()



from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING

from dag_modelling.core.exception import InitializationError
from dag_modelling.core.output import Output
from dag_modelling.parameters import Parameter
from dag_modelling.tools.logger import Logger, get_logger
from numpy import full_like, where

if TYPE_CHECKING:
    from typing import Callable

    from numpy.typing import NDArray


class Minimizable:
    __slots__ = (
        "_statistic",
        "_parameters",
        "_verbose",
        "_functions",
        "_function",
        "_ncall",
        "_nstep",
        "_previous_fcn_step",
        "_previous_parameters_step",
        "_logger",
    )

    _statistic: Output
    _parameters: list[Parameter]
    _verbose: bool
    _functions: dict
    _function: Callable
    _ncall: int
    _nstep: int
    _previous_fcn_step: float | None
    _previous_parameters_step: NDArray | None
    _logger: Logger

    def __init__(
        self,
        statistic: Output,
        parameters: Sequence[Parameter] | None = None,
        verbose: bool = False,
        logger: Logger | None = None,
    ) -> None:
        if not isinstance(statistic, Output):
            raise InitializationError(
                f"'statistic' must be an Output, but given {statistic=}, {type(statistic)=}!"
            )
        self._statistic = statistic

        self._parameters = []  # pyright: ignore
        if parameters is not None:
            if not isinstance(parameters, Sequence):
                raise InitializationError(
                    f"'parameters' must be a sequence of Parameter, but given {parameters=},"
                    f" {type(parameters)=}!"
                )
            for par in parameters:
                self.append_par(par)

        if isinstance(logger, Logger):
            self._logger = logger
        elif logger is None:
            self._logger = get_logger()
        else:
            raise InitializationError(f"Cannot initialize a Minimizable class with logger={logger}")

        self._verbose = verbose
        self._functions = {
            "verbose": self._function_verbose,
            "default": self._function_default,
        }
        self._function = self._functions["verbose" if verbose else "default"]
        self._ncall = 0
        self._nstep = 0
        self._previous_fcn_step = None
        self._previous_parameters_step = None

    def append_par(self, par: Parameter) -> None:
        if not isinstance(par, Parameter):
            raise RuntimeError(f"par must be a Parameter, but given {par=}, {type(par)=}!")
        self._parameters.append(par)

    def _function_default(self, values: NDArray) -> float:
        for param, val in zip(self._parameters, values):
            param.value = val
        self._ncall += 1
        return self._statistic.data[0]

    def _function_verbose(self, values: NDArray) -> float:
        n_parameters = len(self._parameters)
        if self._previous_parameters_step is None:
            parameters_change_step = full_like(values, 0.0)
            modified_parameters_count = 0
            call_type = "initial"
            modified_parameter_numbers = ""
            modified_parameters = ""
        else:
            parameters_change_step = values - self._previous_parameters_step
            modified_parameters_idx = where(parameters_change_step)[0]
            modified_parameters_count = modified_parameters_idx.size

            if n_parameters > 2:
                if modified_parameters_count == n_parameters:
                    call_type = "step"
                elif modified_parameters_count == 1:
                    call_type = "derivative"
                else:
                    call_type = "call type determination failed"
            elif n_parameters == 2:
                if modified_parameters_count == 1:
                    call_type = "derivative"
                else:
                    call_type = "call type determination failed"
            else:
                call_type = "undefined"

            if modified_parameters_idx.size == n_parameters:
                modified_parameter_numbers = "all"
                modified_parameters = "all"

            if modified_parameters_truncated := modified_parameters_count > 2:
                modified_parameters_idx = modified_parameters_idx[:2]

            modified_parameter_numbers = f"{', '.join(map(str, modified_parameters_idx))}"
            modified_parameters = (
                f"{', '.join(self._parameters[i].name for i in modified_parameters_idx)}"
            )
            if modified_parameters_truncated:
                modified_parameter_numbers = f"{modified_parameter_numbers}, ..."
                modified_parameters = f"{modified_parameters}, ..."

        for i, (param, val, change) in enumerate(
            zip(self._parameters, values, parameters_change_step)
        ):
            param.value = val
            if change != 0.0:
                self._logger.info(f"{i: 3d} {param!s} change={change:g}")
            else:
                self._logger.info(f"{i: 3d} {param!s}")

        self._ncall += 1

        ret = self._statistic.data[0]

        self._logger.info(
            f"Modified parameters {modified_parameters_count}/{n_parameters} ({call_type}): {modified_parameter_numbers}; {modified_parameters}"
        )
        self._logger.info(f"Statistic {self._ncall} (step {self._nstep}): {ret}")
        if call_type != "initial":
            fcn_change = ret - self._previous_fcn_step
            self._logger.info(f"Statistic change: {fcn_change:g}")
        self._logger.info("")

        if not call_type == "derivative":
            self._previous_fcn_step = ret
            self._previous_parameters_step = values.copy()
            self._nstep += 1

        return ret

    def __call__(self, values: NDArray) -> float:
        return self._function(values)



from time import perf_counter, time


class FitResult:
    __slots__ = ("_result", "_wall", "_clock")

    _result: dict
    _wall: float
    _clock: float

    def __init__(self):
        self._result = {}

    @property
    def result(self):
        return self._result

    def __enter__(self):
        self._wall = time()
        self._clock = perf_counter()
        return self

    def __exit__(self, *_):
        self._clock = perf_counter() - self._clock
        self._wall = time() - self._wall

    def set(self, x, errors, fun, success, summary, minimizer, nfev, **kwargs):
        result = self._result
        result["fun"] = fun
        result["success"] = success
        result["nfev"] = nfev
        result["minimizer"] = minimizer
        result["clock"] = self._clock
        result["wall"] = self._wall
        result["x"] = x
        result["errors"] = errors
        hess_inv = result["hess_inv"] = kwargs.pop("hess_inv", None)
        if errors is None and hess_inv is not None:
            from numpy import diag  # fmt: skip
            result["errors"] = diag(hess_inv) * 2.0
        result.update(kwargs)
        result["summary"] = summary



from __future__ import annotations

from typing import TYPE_CHECKING

from dag_modelling.core.exception import InitializationError
from dag_modelling.core.output import Output
from dag_modelling.parameters import Parameter
from dag_modelling.tools.logger import Logger, get_logger

from .fit_result import FitResult
from .minimizable import Minimizable

# if we cannot import runtime_error from root we use DagModellingError to avoid any exception capture,
# i.e., if CppRuntimeError==DagModellingError, the exception will be not raised
try:
    import ROOT  # fmt: skip
    CppRuntimeError = ROOT.std.runtime_error
except Exception:
    from dag_modelling.core.exception import DagModellingError  # fmt:skip
    CppRuntimeError = DagModellingError

if TYPE_CHECKING:
    from typing import Any


class MinimizerBase:
    __slots__ = (
        "_name",
        "_label",
        "_minimizable",
        "_minimizer",
        "_parameters",
        "_parameters_names",
        "_result",
        "_statistic",
        "_limits",
        "_verbose",
        "_nbins",
        "_logger",
        "_initial_parameters",
    )

    _name: str
    _label: str
    _minimizable: Minimizable | None
    _parameters: dict[str, Parameter]
    _parameters_names: list[str]
    _result: dict
    _minimizer: Any
    _verbose: bool
    _statistic: Output
    _limits: dict[str, tuple[float | None, float | None]]
    _nbins: int
    _logger: Logger
    _initial_parameters: dict[Parameter, float] | None

    def __init__(
        self,
        statistic: Output,
        parameters: dict[str, Parameter],
        name: str,
        label: str,
        verbose: bool = False,
        logger: Logger | None = None,
        *,
        limits: dict[str, tuple[float | None, float | None]] = {},
        nbins: int = 0,
    ):
        if not isinstance(statistic, Output):
            raise InitializationError(
                f"arg 'statistic' must be an Output, but given {type(statistic)=}, {statistic=}."
            )
        self._statistic = statistic
        self._initial_parameters = {}

        self._parameters = []  # pyright: ignore
        self._parameters_names = []  # pyright: ignore
        if parameters:
            if not isinstance(parameters, dict):
                raise InitializationError(
                    f"'parameters' must be a dict of (str, Parameter), but given {parameters=},"
                    f" {type(parameters)=}!"
                )
            for parameter_name, parameter in parameters.items():
                self.append_name_par(parameter_name, parameter)
                self.copy_initial_values(parameter)

        if isinstance(logger, Logger):
            self._logger = logger
        elif logger is not None:
            raise InitializationError(f"Cannot initialize a Minimizable class with logger={logger}")
        else:
            self._logger = get_logger()

        self._limits = limits

        self._name = name
        self._label = label
        self._verbose = verbose
        self._nbins = nbins
        self._minimizable = None

    @property
    def statistic(self) -> Output:
        return self._statistic

    @property
    def name(self) -> str:
        return self._name

    @property
    def label(self) -> str:
        return self._label

    @property
    def logger(self) -> Logger:
        return self._logger

    @statistic.setter
    def statistic(self, statistic) -> None:
        self._statistic = statistic
        self._minimizable = None

    @property
    def parameters(self) -> list[Parameter]:
        return self._parameters

    @property
    def parameters_names(self) -> list[str]:
        return self._parameters_names

    @property
    def result(self) -> dict:
        return self._result

    @property
    def nbins(self) -> int:
        return self._nbins

    @property
    def npars_free(self) -> int:
        """Return number of free parameters.

        Note
        ----
        It calculates the number of parameters that are considered free at the
        time of initialization. However, if the statistic depends on nuisance
        parameters but does not include pull terms for those parameters, then
        those constrained parameters are not counted as free parameters.

        Returns
        -------
        int
            Number of free parameters.
        """
        return sum(par._parent.is_free for par in self.parameters)

    @property
    def npars_constrained(self) -> int:
        """Return number of constrained parameters.

        Note
        ----
        It calculates the number of parameters that are constrained at the time
        of initialization. However, if the statistic does not include pull terms
        for a parameter, that parameter will be treated as a nuisance parameter.

        Returns
        -------
        int
            Number of constrained parameters.
        """
        return len(self.parameters) - self.npars_free

    def copy_initial_values(self, par: Parameter) -> None:
        self._initial_parameters.update({par: par.value.copy()})

    def push_initial_values(self) -> None:
        for par, value in self._initial_parameters.items():
            par.push(value)

    def append_name_par(self, name: str, par: Parameter) -> None:
        for obj, otype in ((name, str), (par, Parameter)):
            if not isinstance(obj, otype):
                raise RuntimeError(f"'{obj}' must be a {otype}, but given {par=}, {type(par)=}!")
        self._parameters.append(par)
        self._parameters_names.append(name)

    def fit(self, **kwargs) -> dict:
        if len(self.parameters) == 0:
            return self.evalstatistic()
        return self._child_fit(**kwargs)

    def evalstatistic(self) -> dict:
        with FitResult() as fr:
            fun = self._statistic.data

        fr.set(
            x=[],
            errors=[],
            fun=fun,
            success=True,
            summary="stastitics evaluation (no parameters)",
            minimizer="none",
            nfev=1,
        )
        self._result = fr.result
        self.patchresult()

        return self.result

    def _child_fit(self, **_) -> dict:
        raise NotImplementedError("The method must be overriden!")

    def patchresult(self) -> None:
        names = self.parameters_names
        result = self._result
        result["npars"] = len(self.parameters)
        result["names"] = names
        result["xdict"] = dict(zip(names, (float(x) for x in self.result["x"])))
        if self.result["errors"] is not None:
            result["errorsdict"] = dict(zip(names, (float(e) for e in self.result["errors"])))
        else:
            result["errorsdict"] = {}
        result["nbins"] = self.nbins
        result["npars_free"] = self.npars_free
        result["npars_constrained"] = self.npars_constrained
        result["ndof"] = self.nbins - self.npars_free
        result["statistic_name"] = self.statistic.node.name

    def init_minimizable(self) -> Minimizable:
        if self._minimizable is None:
            self._minimizable = Minimizable(
                self.statistic, verbose=self._verbose, logger=self._logger
            )
            for par in self.parameters:
                self._minimizable.append_par(par)
        return self._minimizable
